/****************************************************************
 **
 **  AIRTRAF:
 **     a wireless (802.11) traffic/performance analyzer
 **
 **  connection.c
 **
 ****************************************************************
 **
 **   Copyright (c) 2001 all rights reserved.
 **
 **   Author: Peter K. Lee <pkl@duke.edu>
 **
 ***************************************************************/

/*=============================================================*/
/* System Includes */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <pthread.h>

#include "definition.h"
#include "poll_include.h"

#define MAX_CMD_SIZE 100
#define MAX_QUERY    1000
#define MAX_TEMP     500

extern MYSQL *conn;

/*=============================================================*/
/* Function Definitions */

/**
 * reset_bss_list()
 * ---------------
 * a recursive function to free up all the memory allocated to storing
 * bss_list structures...  
 **/
void reset_bss_list(bss_t * info)
{
  if (info != NULL){
    reset_bss_list(info->next);
    free(info);
    return;
  }
  else{
    return;
  }
}

/**
 * new_bss()
 * --------
 * dynamically initialize new memory to store the info acquired from
 * the sniff server.
 **/
bss_t * new_bss(char * msgbuf)
{
  bss_t * temp;

  fprintf(stderr,"making space... ");
  /** if info's null... **/
  if ((temp = malloc(sizeof(bss_t))) == NULL){
    perror("malloc failed");
    return (NULL);
  }
  //  memset(temp, 0, sizeof(bss_t));
  memcpy(temp, msgbuf, sizeof(bss_t));
  temp->next = NULL;
  fprintf(stderr,"ok\n");
  return (temp);
}

/**
 * dump_traffic_ap()
 * -----------------
 * function to insert ap related info into the database obtained from
 * the sniff server.
 **/
void dump_traffic_ap(bss_t * info, char * address, char * query)
{
  char temp[MAX_TEMP];

  memset(query, 0, MAX_QUERY);
  snprintf(query, MAX_QUERY, "INSERT INTO traffic_ap_5 ");
  strcat(query, "(sniff_server,      ap_bssid,         ap_channel,  ap_ssid,");
  strcat(query, " timestamp,         mgmt_count,       mgmt_byte,   mgmt_beacon,");
  strcat(query, " mgmt_avg_band,     mgmt_high_band,   mgmt_low_band,");
  strcat(query, " mgmt_curr_band,    ctrl_count,       ctrl_byte,   ctrl_ack,");
  strcat(query, " ctrl_avg_band,     ctrl_high_band,   ctrl_low_band,");
  strcat(query, " ctrl_curr_band,    data_count,       data_byte,   data_ext_count,");
  strcat(query, " data_ext_byte,     data_int_count,   data_int_byte,");
  strcat(query, " data_avg_band,     data_high_band,   data_low_band,");
  strcat(query, " data_curr_band,    overall_count,    overall_byte,");
  strcat(query, " overall_corrupt,   overall_nodes,    overall_avg_band,");
  strcat(query, " overall_high_band, overall_low_band, overall_curr_band)");

  snprintf(temp, MAX_TEMP, " VALUES(\"%s\",\"%s\",%d,\"%s\",NOW(), ",
	   address, hexdump((__u8 *)&info->bssid, 6), info->channel,
	   info->ssid);
  strcat(query, temp);
  /** management related data **/
  snprintf(temp, MAX_TEMP, "%d,%d,%d,%-8.4f,%-8.4f,%-8.4f,%-8.4f, ",
	   info->mgmt_data.mgmt_count, info->mgmt_data.mgmt_byte,
	   info->mgmt_data.beacon,
	   info->mgmt_data.bndwth.avg, info->mgmt_data.bndwth.high,
	   info->mgmt_data.bndwth.low, info->mgmt_data.bndwth.curr);
  strcat(query, temp);
  /** control related data **/
  snprintf(temp, MAX_TEMP, "%d,%d,%d,%-8.4f,%-8.4f,%-8.4f,%-8.4f, ",
	   info->ctrl_data.control_count, info->ctrl_data.control_byte,
	   info->ctrl_data.ack,
	   info->ctrl_data.bndwth.avg, info->ctrl_data.bndwth.high,
	   info->ctrl_data.bndwth.low, info->ctrl_data.bndwth.curr);
  strcat(query, temp);
  /** data related data **/
  snprintf(temp, MAX_TEMP, "%d,%d,%d,%d,%d,%d,%-8.4f,%-8.4f,%-8.4f,%-8.4f, ",
	   info->normal_data.data_count, info->normal_data.data_byte,
	   info->normal_data.external_count, info->normal_data.external_byte,
	   info->normal_data.internal_count, info->normal_data.internal_byte,
	   info->normal_data.bndwth.avg, info->normal_data.bndwth.high,
	   info->normal_data.bndwth.low, info->normal_data.bndwth.curr);
  strcat(query, temp);
  /** overall related data  ( !@# - fix corrupt count stuff )**/
  snprintf(temp, MAX_TEMP, "%d,%d,%d,%d,%-8.4f,%-8.4f,%-8.4f,%-8.4f) ",
	   info->overall_count, info->overall_byte,
	   info->overall_count, info->num,
	   info->bndwth.avg, info->bndwth.high,
	   info->bndwth.low, info->bndwth.curr);
  strcat(query, temp);

  /** issue query to insert traffic_ap_5 info **/
  if (db_query(conn,query) < 0){
    fprintf(stderr,"failed\n");
    print_error(conn,"Unable to Insert Table");
    return;
  }
  else{
    fprintf(stderr,"OK\n");
  } 
}

/**
 * dump_traffic_node()
 * -----------------
 * function to insert node related info into the database obtained from
 * the sniff server.
 **/
void dump_traffic_node(bss_t * info, char * query)
{
  char temp[MAX_TEMP];
  char node_mac[50];
  char type[6];
  __u8 bcast[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

  bss_node_t * mynode;
  int c;

  c = 0;
  while (0 != memcmp (&info->addr_list[c].mac_addr,bcast,6))
    {
      mynode = (bss_node_t *)&info->addr_list[c];

      if (0 == memcmp(&mynode->mac_addr,&info->bssid, 6))
	snprintf(type, MAX_TEMP, "AP");
      else
	snprintf(type, MAX_TEMP, "STA");

      snprintf(node_mac, 50 ,"%s", hexdump((__u8 *)&mynode->mac_addr, 6));
      
      memset(query, 0, MAX_QUERY);
      snprintf(query, MAX_QUERY, "INSERT INTO traffic_node_5 ");
      strcat(query, "(ap_bssid, timestamp, node_mac, type, ");
      strcat(query, " incoming_count,  incoming_byte,");
      strcat(query, " outgoing_count,  outgoing_byte,");
      strcat(query, " signal_strength,");
      strcat(query, " node_avg_band,   node_high_band,");
      strcat(query, " node_low_band,   node_curr_band)");

      snprintf(temp, MAX_TEMP, " VALUES(\"%s\",NOW(),\"%s\",\"%s\", ",
	       hexdump((__u8 *)&info->bssid, 6),
	       node_mac,
	       type);
      strcat(query, temp);
      
      snprintf(temp, MAX_TEMP, "%d,%d,%d,%d,%-8.4f,%-8.4f,%-8.4f,%-8.4f,%-8.4f)",
	       mynode->inc_packet, mynode->inc_byte,
	       mynode->out_packet, mynode->out_byte,
	       mynode->avg_signal_str,
	       mynode->bndwth.avg, mynode->bndwth.high,
	       mynode->bndwth.low, mynode->bndwth.curr);
      strcat(query, temp);

      /** issue query to insert traffic_node_5 info **/
      if (db_query(conn,query) < 0){
	fprintf(stderr,"failed\n");
	print_error(conn,"Unable to Insert Table");
	return;
      }
      else{
	fprintf(stderr,"%d... ", (c+1));
      }
      c++;
    }
  fprintf(stderr,"\n%d nodes entered!\n",c);
}

/**
 * dump_to_db()
 * ------------
 * calls appropriate functions to store acquired data into the
 * database.
 **/
void dump_to_db(bss_t * info, char * address)
{
  char query[MAX_QUERY];

  bss_t * tmp = info;
  
  while (tmp != NULL)
    {
      fprintf(stderr,"writing  AP  info into 'traffic_ap_5'   table: ");
      dump_traffic_ap(info, address, query);
      fprintf(stderr,"writing NODE info into 'traffic_node_5' table: ");
      dump_traffic_node(info, query);
      tmp = tmp->next;
    }
}

/**
 * update_sniff_info()
 * -------------------
 * a routine that updates the status of sniff_info, whether it is up,
 * or down, and its last known uptime.
 **/
void update_sniff_info(char * address, char * status)
{
  char query[MAX_QUERY];
  
  if (!strcasecmp("DOWN", status)){
    snprintf(query, MAX_QUERY,
	     "UPDATE sniff_info SET status = \"DOWN\" WHERE ip_address = \"%s\"", address);
  }
  else if (!strcasecmp("UP", status)){
    snprintf(query, MAX_QUERY,
	     "UPDATE sniff_info SET status = \"UP\", last_up = NOW() WHERE ip_address = \"%s\"", address);
  }
  else{
    /** should not get here **/
    return;
  }
  
  /** issue query to update sniff_info table **/
  if (db_query(conn,query) < 0){
    fprintf(stderr,"failed\n");
    print_error(conn,"Unable to Insert Table");
    return;
  }
}


/////////////////////////////////////////////////////////////////
// MAIN CONNECTION ROUTINE
////////////////////////////////////////////////////////////////

/**
 * connections()
 * ------------
 * threaded call to connect to sniff server and retrieve data.
 **/
void *connections(void *addr)
{
  int sockfd;
  struct sockaddr_in servaddr;

  char cmd[MAX_CMD_SIZE];
  char msgbuf[MAX_BUFFER_SIZE * 2];
  char *address;

  bss_t * bss_list = NULL;
  bss_t * temp;
  bss_t * last;
  
  int recvlen;
  int sent;

  int status = 1;

  nthreads++;
  
  /** you can reclaim space when finished... **/
  if (DEBUG) fprintf(stderr,"detaching thread...\n");
  pthread_detach(pthread_self());

  /** get lock so you get exclusive right to address **/
  pthread_mutex_lock(&ready_lock);

  address = (char *)addr;
  if (( sockfd = socket(AF_INET, SOCK_STREAM, 0) ) < 0){
    perror ("Could not create server socket\n");
    status = -1;
  }

  bzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_port = htons(2222);
  if (inet_pton(AF_INET, address, &servaddr.sin_addr)<0){
    perror("cannot resolve IP\n");
    status = -1;
  }

  /** okay used address, let other threads run **/
  pthread_cond_broadcast(&ready_cond);
  pthread_mutex_unlock(&ready_lock);
  
  if (DEBUG) fprintf(stderr,"connecting to host...\n");
  if (connect(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0){
    perror("cannot connect to server\n");
    status = -1;
  }

  /**
   * Put in the status message into the sniff_info table
   * UP||DOWN with last known uptime
   **/
  if (status == -1){
    update_sniff_info(address, "DOWN");
    nthreads--;
    return (NULL);
  }
  else{
    update_sniff_info(address, "UP");
  }
  
  bzero(cmd, MAX_CMD_SIZE);
  switch (command)
    {
    case SYNCH_CMD:
      /** send SYNCH command to sniff server **/
      snprintf(cmd, MAX_CMD_SIZE, "SYNCH");
      if (DEBUG) fprintf(stderr,"synchronizing with sniffer...\n");
      break;
    case GET_DATA_CMD:
      /** send GET DATA command to sniff server **/
      snprintf(cmd, MAX_CMD_SIZE, "GET DATA");
      break;
    }
  if ((sent = send(sockfd, cmd, strlen(cmd), 0)) < 0){
    perror ("Send error\n");
    close(sockfd);
    nthreads--;
    return (NULL);
  }

  if (command == SYNCH_CMD){
    close(sockfd);
    nthreads--;
    return (NULL);
  }
  
  while(1){
    bzero(msgbuf, sizeof(msgbuf));
    if ((recvlen = recv(sockfd,msgbuf,sizeof(bss_t),MSG_WAITALL))<=0){
      perror ("error in receive");
      nthreads--;
      return (NULL);
    }

    fprintf(stderr,"%d bytes of data received\n",recvlen);

    if(strncmp(msgbuf,"\r\n",8) == 0){
      fprintf(stderr,"all data received!\n\n");
      break;
    }

    /** put data into internal structure **/
    if (bss_list == NULL){
      bss_list = new_bss(msgbuf);
    }
    else{
      for(temp = bss_list; temp; temp = temp->next){
	last = temp;
      }
      last->next = new_bss(msgbuf);
    }
  }

  /** dump the collected data into the database **/
  fprintf(stderr,"inserting DATA into database... \n");
  dump_to_db(bss_list, address);
  fprintf(stderr,"all insertion complete!\n\n");
  
  /** clear the bss_list data... **/
  fprintf(stderr,"freeing space... ");
  reset_bss_list(bss_list);
  bss_list = NULL;
  fprintf(stderr,"ok\n");

  /** close the socket **/
  if (close(sockfd) < 0){
    perror ("Close error\n");
  }
  nthreads--;
  return (NULL);
}
