/****************************************************************
 **
 **  AIRTRAF:
 **     a wireless (802.11) traffic/performance analyzer
 **
 **  gui_detailed_scan.c
 **
 ****************************************************************
 **
 **   Copyright (c) 2001, 2002 all rights reserved.
 **
 **   Author: Peter K. Lee <saint@elixar.net>
 **
 ***************************************************************/

#include <ncurses.h>
#include <panel.h>
#include <string.h>
#include <pthread.h>
#include <arpa/inet.h>

#include "definition.h"
#include "sniff_include.h"

extern pthread_mutex_t engine_lock;

/*===========================================================*/
/* Global Variables */

bss_t *bss_list;
int GUI_DEBUG = 0;

bss_t *active_bss = NULL;

bss_t *active_channel[15];
int scroll_loc;
int node_loc;
int node_entry_count;
int max_x, max_y;

////////////////////////////////////////////////////////////////////////////////////
//  Packet Type General Information
////////////////////////////////////////////////////////////////////////////////////

void print_wt_labels(WINDOW * win)
{
    wattrset(win, BOXATTR);
    mvwprintw(win, 2, 2, "BSSID: ");
    mvwprintw(win, 2, 26, "SSID: ");
    mvwprintw(win, 2, 55, "WEP: ");
    mvwprintw(win, 2, 75, "CHANNEL: ");
    mvwprintw(win, 2, COLS - 22, "TIME: ");
    mvwprintw(win, 4, 2, "Management Frames:");
    mvwprintw(win, 12, 2, "Control Frames:");
    mvwprintw(win, 19, 2, "Data Frames:");
    mvwprintw(win, 28, 2, "Corrupt Frames:   (count)  (bytes)");
    mvwprintw(win, 35, 2, "OVERALL ACTIVITY:");
    mvwprintw(win, 40, 2, "Link Quality Analysis:");
    
    wattrset(win, STDATTR);
    /** management **/
    mvwprintw(win, 5,  5, "Beacon:");
    mvwprintw(win, 6,  5, "Disassoc:");
    mvwprintw(win, 7,  5, "Other:");
    mvwprintw(win, 8,  5, "Total Packets:");
    mvwprintw(win, 9,  5, "Total Bytes:");
    wattrset(win, ACTIVEATTR);
    mvwprintw(win, 10, 5, "Bandwidth:");
    /** control **/
    wattrset(win, STDATTR);
    mvwprintw(win, 13, 5, "Acknowledgement:");
    mvwprintw(win, 14, 5, "Other:");
    mvwprintw(win, 15, 5, "Total Packets:");
    mvwprintw(win, 16, 5, "Total Bytes:");
    wattrset(win, ACTIVEATTR);
    mvwprintw(win, 17, 5, "Bandwidth:");
    /** data **/
    wattrset(win, STDATTR);
    mvwprintw(win, 20, 5, "External Packets:");
    mvwprintw(win, 21, 5, "External Bytes:");
    mvwprintw(win, 22, 5, "Internal Packets:");
    mvwprintw(win, 23, 5, "Internal Bytes:");
    mvwprintw(win, 24, 5, "Total Packets:");
    mvwprintw(win, 25, 5, "Total Bytes:");
    wattrset(win, ACTIVEATTR);
    mvwprintw(win, 26, 5, "Bandwidth:"); 
    /** corrupted **/
    wattrset(win, STDATTR);
    mvwprintw(win, 29, 5, "Bad MAC addr:");
    mvwprintw(win, 30, 5, "Bad IP chksum:");
    mvwprintw(win, 31, 5, "FCS error:");
    mvwprintw(win, 32, 5, "Filtered data:");
    wattrset(win, ACTIVEATTR);
    mvwprintw(win, 33, 5, "Overall:");
    /** OVERALL **/
    wattrset(win, STDATTR);
    mvwprintw(win, 36, 5, "Total Packets:");
    mvwprintw(win, 37, 5, "Total Bytes:");
    wattrset(win, ACTIVEATTR);
    mvwprintw(win, 38, 5, "Bandwidth:"); 
    /** Link Quality **/
    wattrset(win, STDATTR);
    mvwprintw(win, 41, 5, "Link Utilization:");
    mvwprintw(win, 42, 5, "Background Noise:");
    mvwprintw(win, 43, 5, "Packet Loss:");
    
    update_panels();
    doupdate();
}
void printstatrow(WINDOW * win, int row, int col, unsigned long long number)
{
  wmove(win,row,col);
  printlargenum(number, win);
}

void printdetails(WINDOW * win, bss_t *info)
{
  detailed_overview_t *overview;
  
  wattrset(win, HIGHATTR);
  if (info != NULL){
    mvwprintw(win, 2, 10, hexdump((__u8*)&info->bssid,6));
    mvwprintw(win, 2, 33, info->ssid);
    mvwprintw(win, 2, 61, "%s", info->wep_status ? "encrypted " : "opensystem");
    mvwprintw(win, 2, 85, "%d", info->channel);
    
    /** management **/
    printstatrow(win, 5, 21, info->mgmt_data.beacon);
    printstatrow(win, 6, 21, info->mgmt_data.disassoc);
    printstatrow(win, 7, 21, info->mgmt_data.other);
    printstatrow(win, 8, 21, info->mgmt_data.mgmt_count);
    printstatrow(win, 9, 21, info->mgmt_data.mgmt_byte);
    mvwprintw(win,10, 22, "%8.2f Kbps", info->mgmt_data.bndwth.curr);
    /** control **/
    printstatrow(win,13, 21, info->ctrl_data.ack);
    printstatrow(win,14, 21, info->ctrl_data.other);
    printstatrow(win,15, 21, info->ctrl_data.control_count);
    printstatrow(win,16, 21, info->ctrl_data.control_byte);
    mvwprintw(win,17, 22, "%8.2f Kbps", info->ctrl_data.bndwth.curr);
    /** data **/
    printstatrow(win, 20, 21, info->normal_data.external_count);
    printstatrow(win, 21, 21, info->normal_data.external_byte);
    printstatrow(win, 22, 21, info->normal_data.internal_count);
    printstatrow(win, 23, 21, info->normal_data.internal_byte);
    printstatrow(win, 24, 21, info->normal_data.data_count);
    printstatrow(win, 25, 21, info->normal_data.data_byte);
    mvwprintw(win,26, 22, "%8.4f Mbps", info->normal_data.bndwth.curr);
    /** corrupt **/
    overview = get_detailed_snapshot();
    mvwprintw(win,29, 21, "%5d  %7d", overview->bad_mac,
	      overview->bad_mac_byte);
    mvwprintw(win,30, 21, "%5d  %7d", overview->bad_ip_chksum,
	      overview->bad_ip_chksum_byte);
    mvwprintw(win,31, 21, "%5d  %7d", overview->fcs_error,
	      overview->fcs_error_byte);
    mvwprintw(win,32, 21, "%5d  %7d", overview->filtered_data,
	      overview->filtered_data_byte);
    mvwprintw(win,33, 21, "%5d  %7d", overview->corrupt_tot,
	      overview->corrupt_tot_byte);
    	      
    /** overall **/
    printstatrow(win, 36, 21, info->overall_count);
    printstatrow(win, 37, 21, info->overall_byte);
    mvwprintw(win,38, 22, "%8.4f Mbps", info->bndwth.curr);

    mvwprintw(win,41, 23, "%7.2f %%", info->link_utilization);
    mvwprintw(win,42, 23, "%7.2f %%", info->background_noise);
    mvwprintw(win,43, 23, "%7.2f %%", info->packet_loss);
  }
}

void markactive(int curwin, WINDOW * tw, int paused)
{
    WINDOW *win1;
    int x1, y1;

    win1 = tw;

    getmaxyx(win1, y1, x1);

    if (!paused){
      wmove(win1, --y1, x1 - 10);
      wattrset(win1, ACTIVEATTR);
      wprintw(win1, " Active ");
    }
    else{
      wmove(win1, --y1, x1- 10);
      wattrset(win1, ACTIVEATTR);
      wprintw(win1, " Paused ");
    }

}

////////////////////////////////////////////////////////
//  Scroller & Connected Nodes Stuff
////////////////////////////////////////////////////////

void prepare_node_win(WINDOW *nodewin)
{
  scroll_loc = node_loc = node_entry_count = 0;

  getmaxyx(nodewin, max_y, max_x);  
}

void count_node()
{
  __u8 bcast[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
  
  bss_t * curr_list = bss_list;
  int c = 0;

  if (bss_list == NULL){
    return;
  }
  while (0 != memcmp (&curr_list->addr_list[c].mac_addr,bcast,6)){
    c++;
  }
  node_entry_count = c;
}

void show_node_win(bss_t * info, WINDOW *nodewin)
{
  char address[20];
  int i;
  int count = 0;
  int index = scroll_loc;
  int pos;
  int fit, remainder;

  fit = max_y / NODE_ROW_SIZE;
  remainder = max_y % NODE_ROW_SIZE;

  if ((node_entry_count == 0) && (info == NULL)){
    return;
  }
  
  for(i = index; i < node_entry_count; i++){
    if (!info->addr_list[i].status){
      continue;
    }
    pos = (count * NODE_ROW_SIZE) +1;
    wattrset(nodewin,BOXATTR);
    if (info->addr_list[i].ip_addr.s_addr != 0){
      snprintf(address, 20, "%s",
	       inet_ntoa((struct in_addr)info->addr_list[i].ip_addr));
    }
    else{
      snprintf(address,20,"Unknown");
    }
    mvwprintw(nodewin, pos, 0,
	      "MAC address %d:  %s -       IP: (%s) ", i,
	      hexdump((__u8 *)&info->addr_list[i].mac_addr,6),
	      address);
    
    wattrset(nodewin,ACTIVEATTR);
    if (0 == memcmp(&info->addr_list[i].mac_addr,
		    &info->bssid, 6))
      mvwprintw(nodewin, pos, 30, " AP");
    else
      mvwprintw(nodewin, pos, 30, " STA");
    wattrset(nodewin,STDATTR);
    mvwprintw(nodewin, pos+1, 5,
	      "incoming packets:                outgoing packets: ");
    mvwprintw(nodewin, pos+2, 5,
	      "incoming bytes:                  outgoing bytes: ");
    mvwprintw(nodewin, pos+3, 5, "avg.signal strength: ");
    wattrset(nodewin, ACTIVEATTR);
    mvwprintw(nodewin, pos+4, 5, "Bandwidth: ");

    wattrset(nodewin,HIGHATTR);
    printstatrow(nodewin, pos+1, 25, info->addr_list[i].inc_packet);
    printstatrow(nodewin, pos+1, 56, info->addr_list[i].out_packet);
    printstatrow(nodewin, pos+2, 25, info->addr_list[i].inc_byte);
    printstatrow(nodewin, pos+2, 56, info->addr_list[i].out_byte);
    mvwprintw(nodewin, pos+3, 29, "%-10.2f", info->addr_list[i].avg_signal_str);
    mvwprintw(nodewin, pos+4, 29, "%-8.4f Mbps", info->addr_list[i].bndwth.curr);
    count++;
    if (count == fit){
      break;
    }
  }
}

void scroll_nodewin(WINDOW *nodewin, int direction)
{
  int temp;

  wattrset(nodewin, STDATTR);

  if (direction == SCROLLUP) {
    if (scroll_loc > 0){
      scroll_loc--;
    }
    else{
      scroll_loc = 0;
    }
  }
  else {
    temp = (node_entry_count * NODE_ROW_SIZE) - max_y;
    if (temp < 0){
      scroll_loc = 0;
    }
    else{
      temp = (temp/NODE_ROW_SIZE) + 1;
      if (scroll_loc < temp){
	scroll_loc++;
      }
      else{
	scroll_loc = temp;
      }
    }
  }
  show_node_win(active_bss, nodewin);
}

/* void page_nodewin(WINDOW *nodewin, int direction) */
/* { */
/*     int i = 1; */

/*     if (direction == SCROLLUP) { */
/* 	while ((i <= LINES - 7) && (table->lastvisible != table->tail)) { */
/* 	    i++; */
/* 	    scrollethwin(table, direction, idx); */
/* 	} */
/*     } else { */
/* 	while ((i <= LINES - 7) && (table->firstvisible != table->head)) { */
/* 	    i++; */
/* 	    scrollethwin(table, direction, idx); */
/* 	} */
/*     } */
/* } */

/////////////////////////////////////////////////////////////////////
//  CHANNEL Panel STUFF
/////////////////////////////////////////////////////////////////////

void prepare_channel_win(WINDOW * win)
{
  int i;
  int loc = 17;
  
  wattrset(win, IPSTATLABELATTR);
  wmove(win, 0, 1);
  wprintw(win, "CHANNEL STATUS:");

  for(i = 1; i < 15; i++){
    mvwprintw(win, 0, loc, "%d", i);
    if (i <10){
      loc += 3;
    }
    else{
      loc +=4;
    }
  }
}

/** highlight available channels for viewing **/
void show_channel_stats(WINDOW * win)
{
  int i;
  int loc = 17;
  
  wattrset(win, ERRTXTATTR);

  for(i = 1; i < 15; i++){
    if (active_channel[i] != NULL){
      mvwprintw(win, 0, loc, "%d", i);
    }
    if (i < 10){
      loc += 3;      
    }
    else{
      loc += 4;
    }
  }
}

void find_active_channel(bss_t *info)
{
  bss_t * curr_info = info;

  while(curr_info)
    {
      if (curr_info->channel > 0){
	active_channel[curr_info->channel] = curr_info;
      }
      curr_info = curr_info->next;
    }
}

void goto_next_channel()
{
  int curr_chan;
  int i;
  
  if (active_bss != NULL){
    curr_chan = active_bss->channel;
  }
  else{
    return;
  }

  for(i = (curr_chan + 1); i < 15; i++){
    if (active_channel[i] != 0){
      active_bss = (bss_t *)active_channel[i];
      return;
    }
  }
  for(i = 1; i < curr_chan; i++){
    if (active_channel[i] != 0){
      active_bss = (bss_t *)active_channel[i];
      return;
    }
  }
}

void goto_prev_channel()
{
  int curr_chan;
  int i;
  
  if (active_bss != NULL){
    curr_chan = active_bss->channel;
  }
  else{
    return;
  }

  for(i = (curr_chan-1); i > 0; i--){
    if (active_channel[i] != 0){
      active_bss = (bss_t *)active_channel[i];
      return;
    }
  }
  for(i = 14; i > curr_chan; i--){
    if (active_channel[i] != 0){
      active_bss = (bss_t *)active_channel[i];
      return;
    }
  }
}


////////////////////////////////////////////////////////////////////
//  main wireless traffic monitor routine
///////////////////////////////////////////////////////////////////

int start_wt_mon(struct SETTINGS *mySettings)
{
  WINDOW *statwin;
  WINDOW *nodeborder;
  WINDOW *nodewin;
  WINDOW *channelwin;
    
  PANEL *statpanel;
  PANEL *nodeborderpanel;
  PANEL *nodepanel;
  PANEL *channelpanel;

  int exitloop;

  int ch;

  int first_instance = 1;
  int paused = 0;  // used to pause screen
    
  int curwin = 0;
  
  struct timeval tv_start;
  struct timeval tv_curr;
  struct timeval tv_pot_reset;
  struct timeval tv_old;
  struct timeval tv_new;

  char elapsed[10];

  char sp_buf[10];

  if (GUI_DEBUG) fprintf(stderr,"making overall window\n");
  /** make the overall wireless traffic monitor box window **/
  statwin = newwin(LINES - 3, COLS, 1, 0);
  statpanel = new_panel(statwin);
  stdwinset(statwin);
  wtimeout(statwin, -1);
  wattrset(statwin, BOXATTR);
  colorwin(statwin);
  box(statwin, ACS_VLINE, ACS_HLINE);
  wmove(statwin, 0, 1);
  wprintw(statwin, " Statistics for %s ", mySettings->interface);
  wattrset(statwin, STDATTR);
  print_wt_labels(statwin);
  printdetails(statwin, NULL);
  update_panels();
  doupdate();

  if (GUI_DEBUG) fprintf(stderr,"making node window\n");
  /** make the node connection traffic monitor box window **/
  nodeborder = newwin(LINES - 8, COLS - 40, 5, 39);
  nodeborderpanel = new_panel(nodeborder);
  nodewin = newwin(LINES - 10, COLS - 42, 6, 40);
  nodepanel = new_panel(nodewin);
  wattrset(nodeborder, BOXATTR);
  box(nodeborder, ACS_VLINE, ACS_HLINE);
  wmove(nodeborder, 0, 1);
  wprintw(nodeborder, " Connected Nodes ");

  stdwinset(nodewin);
  wtimeout(nodewin, -1);
  wattrset(nodewin, STDATTR);
  colorwin(nodewin);
  prepare_node_win(nodewin);
  show_node_win(NULL, nodewin);
  update_panels();
  doupdate();

  if (GUI_DEBUG) fprintf(stderr,"making channel window\n");
  /** make the channel display info (shows active channels) **/
  channelwin = newwin(1, COLS, LINES - 2, 0);
  channelpanel = new_panel(channelwin);
  wattrset(channelwin, IPSTATLABELATTR);
  wmove(channelwin, 0, 0);
  sprintf(sp_buf, "%%%dc", COLS);
  wprintw(channelwin, sp_buf, ' ');
  prepare_channel_win(channelwin);
  show_channel_stats(channelwin);

  if (GUI_DEBUG) fprintf(stderr,"making other stuff\n");
  /** make the bottom key binding panel **/
  move(LINES - 1, 1);
  scrollkeyhelp();
  channelkeyhelp();
  pausekeyhelp();
  stdexitkeyhelp();

  markactive(curwin, nodeborder, paused);
  update_panels();
  doupdate();

  gettimeofday(&tv_start, NULL);
  tv_old = tv_start;
  tv_pot_reset = tv_start;
    
  leaveok(statwin, TRUE);
   
  exitloop = 0;
  //    dispmode(options->actmode, unitstring);

  /*
   * Data Display Loop
   */
  while ((!exitloop) && (!sysexit)) {
    ch = ERR;
    if (check_for_keystroke() != ERR)
      ch = wgetch(statwin);
    if (ch != ERR) {
      switch (ch)
	{
	case KEY_UP:
	  scroll_nodewin(nodewin, SCROLLUP);
	  break;
	case KEY_DOWN:
	  scroll_nodewin(nodewin, SCROLLDOWN);
	  break;
	case KEY_PPAGE:
	case '-':
	  scroll_nodewin(nodewin, SCROLLUP);
	  //	    pageethwin(nodewin, SCROLLDOWN);
	  break;
	case KEY_NPAGE:
	case ' ':
	  scroll_nodewin(nodewin, SCROLLDOWN);
	  //	    pageethwin(nodewin, SCROLLUP);
	  break;
	case KEY_RIGHT:
	  goto_next_channel();
	  break;
	case KEY_LEFT:
	  goto_prev_channel();
	  break;
	case 'p':
	case 'P':
	  switch (paused)
	    {
	    case 0:
	      paused = 1;
	      break;
	    case 1:
	      paused = 0;
	      break;
	    }    
	  markactive(curwin, nodeborder, paused);
	  update_panels();
	  doupdate();
	  break;
	case 12:
	case 'l':
	case 'L':
	  refresh_screen();
	  break;
	case 'c':
	case 'C':
	  bss_list = NULL;
	  break;
	case 'Q':
	case 'q':
	case 'X':
	case 'x':
	case 24:
	case 27:
	  exitloop = 1;
	  break;
	default:
	  break;
	}
    }
    if(!paused){
      /** get engine lock before displaying stuff **/
      pthread_mutex_lock(&engine_lock);

      gettimeofday(&tv_new, NULL);

      if (GUI_DEBUG) fprintf(stderr,"doing packet stuff\n");

      /**
       * if it is first instance, and we haven't found a base
       * station yet, then just loop on, waiting for perhaps some
       * keystroke that will find a base channel for us...
       **/
      if (first_instance){
	if (get_detailed_snapshot()->bss_list_top != NULL){
	  bss_list = get_detailed_snapshot()->bss_list_top;
	  active_bss = bss_list;
	  first_instance = 0;
	}
	else{
	  pthread_mutex_unlock(&engine_lock);
	  continue;
	}
      }

      /**
       * since we found a base station, lets see if there are more
       * than one...
       **/
      find_active_channel(bss_list);
      show_channel_stats(channelwin);
      if (active_bss == NULL){
	active_bss = bss_list;
      }
      printdetails(statwin, active_bss);
      count_node();
      show_node_win(active_bss, nodewin);
      update_panels();
      doupdate();

      wattrset(statwin, HIGHATTR);
      gettimeofday(&tv_curr, NULL);
      get_elapsed_time(&tv_curr, &tv_start, elapsed);
      mvwprintw(statwin, 2, COLS - 15, "%s", elapsed);
	
      /** release engine lock **/
      pthread_mutex_unlock(&engine_lock);
    }
  } // end loop
  del_panel(channelpanel);
  delwin(channelwin);
  del_panel(nodeborderpanel);
  delwin(nodeborder);
  del_panel(nodepanel);
  delwin(nodewin);
  del_panel(statpanel);
  delwin(statwin);
  update_panels();
  doupdate();
  if (sysexit){
    sysexit = 0;
    return (0);
  }
  else{
    return (1);
  }
}
