/****************************************************************
 **
 **  AIRTRAF:
 **     a wireless (802.11) traffic/performance analyzer
 **
 **  gui_main.c
 **
 ****************************************************************
 **
 **   Copyright (c) 2001,2002 all rights reserved.
 **
 **   Author: Peter K. Lee <pkl@duke.edu>
 **
 ***************************************************************/

/*=============================================================*/
/* System Includes */

#include <ncurses.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <asm/types.h>
#include <sys/types.h>
#include <sys/time.h>
#include <net/if.h>
#include <linux/netlink.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <errno.h>
#include <pthread.h>

#include "definition.h"
#include "sniff_include.h"

///////////////////////////////////////////////////////////////
//  HELPER FUNCTIONS
///////////////////////////////////////////////////////////////

/**
n * check_for_keystroke()
 * --------------------
 * the helper function for all the GUI interface that checks if there
 * are keystrokes to be read and processed
 **/
int check_for_keystroke()
{
  fd_set fds;
  int ret_val;
  struct timeval tv;
  
  /**
   * a litle loop to check for keystrokes in INTERACTIVE MODE
   **/
  FD_ZERO(&fds);
  FD_SET(0,&fds);
  tv.tv_sec = 0;
  tv.tv_usec = DEFAULT_UPDATE_DELAY;
  
  do {
    ret_val = select(1, &fds, NULL, NULL, &tv);
  } while ((ret_val < 0) && (errno == EINTR));
  
  if (FD_ISSET(0, &fds)){
    return (1);
  }
  else{
    return (ERR);
  }
}

/**
 * change_channel()
 * ---------------
 * A simple routine for changing the channel the wireless card is
 * sniffing on (a necessity for prismII chipset cards)
 * returns: a new socket handler
 **/
int change_channel(int channel)
{
  if ( (channel < 1) || (channel > 14) ){
    return (ERR);
  }
  /** do some driver level channel setting function... **/

  /** then driver level new socket creation, returning socket file
      descriptor **/
  
  return (OK);
}

/**
 * select_specified_ap()
 * ---------------------
 * display menu with all discovered access point, allows user to
 * select a specified access point to listen on.
 **/
void select_specified_ap(struct MENU *menu, struct access_point **chosen, int *abort)
{
  int i;
  int row;

  int endloop = 0;
  int aborted;
  int index = 0;
  int ap_pos;

  char holder[100];
  char description[200];
  
  struct access_point *ap;
  struct channel_overview *snapshot;

  snapshot = get_channel_snapshot();
  
  /** shouldn't happen...**/
  if (snapshot == NULL){
    *abort = 1;
    return;
  }
  
  initmenu(menu, snapshot->num_det_aps + 6, 55,
	     10, (COLS - 55) / 2);

  ap_pos = 0;
  for (i = 1; i < 15; i++){
    if ((ap = snapshot->all_chan[i]) != NULL){
      while (ap != NULL){
	ap_pos++;
	memset(holder, 0, sizeof(char)*100);
	snprintf(holder, strlen((char*)ap->ssid) + 13,
		 " ^%d^ %s CH:%02d",ap_pos, ap->ssid, ap->channel);
	memset(description, 0, sizeof(char)*200);
	snprintf(description, 200,
		 "Select to view data for this access point, WEP Status: %s",
		 ap->wep_status ? "encrypted" : "opensystem");
	additem(menu, holder, description);
	ap = (struct access_point*)ap->next;
      }
    }
  }  
  additem(menu, NULL, NULL);
  additem(menu, " Dynamic ^d^iscovery",
	  "Perform dynamic discovery of access points (not recommended when amplified)");
  additem(menu, NULL, NULL);
  additem(menu, " E^x^it menu", "Exit this menu");
  
  row = 1;
  do {
    showmenu(menu);
    operatemenu(menu, &row, &aborted);
    if (row <= snapshot->num_det_aps){
      index = row;
      endloop = 1;
    }
    else if (row == snapshot->num_det_aps + 2){
      *chosen = NULL;
      endloop = 1;      
    }
    else{
      *abort = 1;
      endloop = 1;
    }
  } while (!endloop);
  
  if (index != 0){
    ap_pos = 0;
    for (i = 1; i < 15; i++){
      if ((ap = snapshot->all_chan[i]) != NULL){
	while (ap != NULL){
	  ap_pos++;
	  if (ap_pos == index){
	    *chosen = ap;
	    return;
	  }
	  ap = (struct access_point*)ap->next;	  
	}
      }
    }
  }
}

void print_selected_ap_screen(WINDOW **win, PANEL **panel)
{
  *win = newwin(8, 35, 5, (COLS - 35) / 2);
  *panel = new_panel(*win);
  wattrset(*win, BOXATTR);
  colorwin(*win);
  box(*win, ACS_VLINE, ACS_HLINE);

  wattrset(*win, ACTIVEATTR);
  mvwprintw(*win, 1, 3, "Current Target Access Point");
  wattrset(*win, STDATTR);
  mvwprintw(*win, 3, 3, "SSID:             ");
  mvwprintw(*win, 4, 3, "BSSID: ");
  mvwprintw(*win, 5, 3, "WEP: ");
  mvwprintw(*win, 6, 3, "Channel:");

  update_panels();
  doupdate();
}

void print_selected_ap_results(WINDOW *win, PANEL *panel, struct access_point *ap)
{
  wattrset(win, HIGHATTR);
  if (ap == NULL){
    mvwprintw(win, 3, 10, "   --              ");
    mvwprintw(win, 4, 10, "   --              ");
    mvwprintw(win, 5, 10, "   --              ");
    mvwprintw(win, 6, 13, "--              ");
  }
  else{
    mvwprintw(win, 3, 10, "                    ");
    mvwprintw(win, 4, 10, "                    ");
    mvwprintw(win, 6, 13, "      ");       
    mvwprintw(win, 3, 10, ap->ssid);
    mvwprintw(win, 4, 10, hexdump((__u8*)&ap->bssid,6));
    mvwprintw(win, 5, 10, "%s", ap->wep_status ? "encrypted " : "opensystem");
    mvwprintw(win, 6, 13, "%02d", ap->channel);
  }
  update_panels();
  doupdate();
}

/**
 * program_interface()
 * ------------------
 * Internal program handler for the entire ncurses GUI
 **/
void program_interface(struct SETTINGS *mySettings)
{
  struct MENU menu;
  struct MENU ap_select;

  WINDOW * selected_ap_win;
  PANEL * selected_ap_panel;
  
  struct access_point *chosen = NULL;

  int channel_scanned = 0;
  int endloop = 0;
  int row = 1;

  int abort;
  int resp;
  int aborted;
  
  draw_desktop();
   
  attrset(STATUSBARATTR);
  mvprintw(0, 1, "AirTraf: %s '02", VERSION_INFO);
    
  attrset(STATUSBARATTR);
  mvprintw(LINES - 1, 1, "LINUX! - For any questions, contact me at saint@elixar.net");
  about();
  
  initmenu(&menu, 17, 35, (LINES - 16) / 2, (COLS - 35) / 2);
  
  additem(&menu, " ^S^can Channels for AP Activity",
	  "Displays list of ALL available access points");
  additem(&menu, " ^C^hange selected target AP",
	  "Allows you to select the access point you wish to analyze");
  additem(&menu, NULL, NULL);
  additem(&menu, " Detailed ^A^ccess Point monitor",
	  "Displays detailed wireless traffic information");
  additem(&menu, " ^G^eneral protocol statistics",
	  "Displays general protocol breakdown (MAC, Network, Transport)");
  additem(&menu, " TCP ^P^erformance Analysis",
	  "Displays detailed analysis of TCP performance");
  additem(&menu, " Service ^B^reakdown by Port",
	  "Displays service usage breakdown for TCP & UDP connections");
  additem(&menu, " Intrusion ^d^etection statistics",
	  "Displays statistics for unauthorized activity (broken)");
  additem(&menu, " ^T^est access point security",
	  "Displays results of security check on specified AP (not yet available)");
  additem(&menu, NULL, NULL);
  additem(&menu, " ^F^ilters...",
	  "Allows you to select traffic display and logging criteria (coming soon!)");
  additem(&menu, NULL, NULL);
  additem(&menu, " C^o^nfigure...", "Set various program options (coming soon!)");
  additem(&menu, NULL, NULL);
  additem(&menu, " E^x^it", "Exits program");

  endloop = 0;

  print_selected_ap_screen(&selected_ap_win, &selected_ap_panel);
  print_selected_ap_results(selected_ap_win, selected_ap_panel, chosen);
  
  do {
    showmenu(&menu);
    operatemenu(&menu, &row, &aborted);

    switch (row) {
    case 1:
      mySettings->scan_mode = CHANNEL_SCAN;
      initialize_channel_scan();
      start_sniffer_engine(mySettings);
      if (start_ap_mon(mySettings))
	stop_sniffer_engine(mySettings);
      channel_scanned = 1;
      /** as soon as we get out, have user pick access point **/
      select_specified_ap(&ap_select, &chosen, &abort);
      destroymenu(&ap_select);
      print_selected_ap_results(selected_ap_win, selected_ap_panel, chosen);
      if (abort){
	break;
      }
      break;
    case 2:
      if (!channel_scanned){
	errbox("Channels have not yet been scanned...  Please scan channels first!",
	       ANYKEY_MSG, &resp);
	break;
      }
      select_specified_ap(&ap_select, &chosen, &abort);
      destroymenu(&ap_select);
      print_selected_ap_results(selected_ap_win, selected_ap_panel, chosen);
      if (abort){
	break;
      }     
    case 4:
      if (!channel_scanned){
	errbox("Channels have not yet been scanned...  Please scan channels first!",
	       ANYKEY_MSG, &resp);
      }
      else{
	if (get_channel_snapshot()->num_det_aps < 1){
	  errbox("No Activty found on any channels, cannot perform detailed scan!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	if (chosen == NULL){
	  errbox("You have not chosen access point yet!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	/** no need for locks & alerts since engine not running here **/
	if (mySettings->card_type == PRISMII){
	  select_channel(mySettings, chosen->channel);
	}
	mySettings->scan_mode = DETAILED_SCAN;
	mySettings->chosen_ap = (void *)chosen;
	initialize_detailed_scan();
	init_potential_structs();
	start_sniffer_engine(mySettings);
	if (start_wt_mon(mySettings))
	  stop_sniffer_engine(mySettings);
	free_potential_structs();
	free_detailed_scan();
      }
      break;
    case 5:
      if (!channel_scanned){
	errbox("Channels have not yet been scanned...  Please scan channels first!",
	       ANYKEY_MSG, &resp);
      }
      else{
	if (get_channel_snapshot()->num_det_aps < 1){
	  errbox("No Activty found on any channels, cannot perform detailed scan!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	if (chosen == NULL){
	  errbox("You have not chosen access point yet!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	/** no need for locks & alerts since engine not running here **/
	if (mySettings->card_type == PRISMII){
	  select_channel(mySettings, chosen->channel);
	}
	mySettings->scan_mode = DETAILED_SCAN;
	mySettings->chosen_ap = (void *)chosen;
	initialize_detailed_scan();
	init_potential_structs();
	start_sniffer_engine(mySettings);
	if (start_gen_proto_mon(mySettings))
	  stop_sniffer_engine(mySettings);
	free_potential_structs();
	free_detailed_scan();
      }
      break;
    case 6:
      if (!channel_scanned){
	errbox("Channels have not yet been scanned...  Please scan channels first!",
	       ANYKEY_MSG, &resp);
      }
      else{
	if (get_channel_snapshot()->num_det_aps < 1){
	  errbox("No Activty found on any channels, cannot perform detailed scan!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	if (chosen == NULL){
	  errbox("You have not chosen access point yet!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	/** no need for locks & alerts since engine not running here **/
	if (mySettings->card_type == PRISMII){
	  select_channel(mySettings, chosen->channel);
	}
	mySettings->scan_mode = DETAILED_SCAN;
	mySettings->chosen_ap = (void *)chosen;
	initialize_detailed_scan();
	init_potential_structs();
	start_sniffer_engine(mySettings);
	if (start_tcp_analysis_mon(mySettings))
	  stop_sniffer_engine(mySettings);
	free_potential_structs();
	free_detailed_scan();
      }
      break;
    case 7:
      if (!channel_scanned){
	errbox("Channels have not yet been scanned...  Please scan channels first!",
	       ANYKEY_MSG, &resp);
      }
      else{
	if (get_channel_snapshot()->num_det_aps < 1){
	  errbox("No Activty found on any channels, cannot perform detailed scan!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	if (chosen == NULL){
	  errbox("You have not chosen access point yet!",
		 ANYKEY_MSG, &resp);
	  break;
	}
	/** no need for locks & alerts since engine not running here **/
	if (mySettings->card_type == PRISMII){
	  select_channel(mySettings, chosen->channel);
	}
	mySettings->scan_mode = DETAILED_SCAN;
	mySettings->chosen_ap = (void *)chosen;
	initialize_detailed_scan();
	init_potential_structs();
	start_sniffer_engine(mySettings);
	if (start_gen_proto_mon(mySettings))
	  stop_sniffer_engine(mySettings);
	free_potential_structs();
	free_detailed_scan();
      }
      break;
    case 8:
      errbox("Closed for re-modeling... :)",
	     ANYKEY_MSG, &resp);
      //      start_sniffer_engine(mySettings);
      //      start_ids_mon(mySettings->interface);
      //      stop_sniffer_engine(mySettings);
      break;
    case 9:
      errbox("Coming very very soon!!!",
	     ANYKEY_MSG, &resp);
      break;
    case 10:
      //config_filters(&ofilter);
      //savefilters(&ofilter);
      break;
    case 13:
      //setoptions(options, &ports);
      //saveoptions(options);
      break;
    case 15:
      endloop = 1;
      break;
    }  
  } while ((!endloop) && (!sysexit));
  del_panel(selected_ap_panel);
  delwin(selected_ap_win);
  destroymenu(&menu);
  erase();
  update_panels();
  doupdate();
}
  
